# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import copy

import pysensu_yelp


class SensuTTLAlerter(object):
    """ This class triggers a sensu alert if the producer dies.  If the sensu_event_info is
    not sent within the ttl specified the prior time it was sent then sensu will alert.

    Args:
        sensu_event_info(dict): dictionary with sensu parameters.  For details see
             http://pysensu-yelp.readthedocs.io/en/latest/index.html?highlight=send_event
             for details
        enable(bool): enable this ttl alert manager
    """

    def __init__(self, sensu_event_info, enable=True):
        self._sensu_event_info = sensu_event_info
        self._enable = enable

    def process(self):
        if self.enable:
            pysensu_yelp.send_event(**self._sensu_event_info)

    @property
    def enable(self):
        return self._enable

    @enable.setter
    def enable(self, new_enable_value):
        if self._enable and not new_enable_value:
            # send final message without ttl
            final_sensu_info = copy.deepcopy(self._sensu_event_info)
            final_sensu_info.pop('ttl')
            pysensu_yelp.send_event(**final_sensu_info)
        self._enable = new_enable_value



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import re
import signal
import time
from datetime import datetime
from optparse import OptionGroup

from cached_property import cached_property
from yelp_batch import Batch
from yelp_batch import batch_command_line_options
from yelp_batch import batch_configure
from yelp_batch._db import BatchDBMixin
from yelp_conn.connection_set import ConnectionDef
from yelp_conn.connection_set import ConnectionSet
from yelp_conn.sqlatxn import TransactionManager
from yelp_conn.topology import ConnectionSetConfig
from yelp_conn.topology import TopologyFile

from data_pipeline import __version__
from data_pipeline.schematizer_clientlib.models.refresh import RefreshStatus
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer
from data_pipeline.servlib.config_util import load_default_config


# WARNING: This batch will not work for non-numeric or compound primary
# keys cause of the logic in process_table(). There is another way of
# doing the same but its much slower and resource intensive.
# https://goo.gl/Drz0uj
class FullRefreshRunner(Batch, BatchDBMixin):
    """The FullRefreshManager spawns off a child process which executes this
    batch to copy rows from the source table into a black hole table.

    Args:
        refresh_id (int): Identifies the refresh to be executed.
        cluster (str): The cluster name of the table to be refreshed.
        database (str): The database name that the table belongs to.
        ro_replica (str): The ro replica connection name for this cluster.
        rw_replica (str): The rw replica connection name for this cluster.
        config_path (str): Path to the config file containing information
            to initialize yelp_conn.
        table_name (str): Name of the table to be refreshed.
        offset (int): The row number to start the refresh from.
        batch_size (int): The number of rows to refresh per batch.
        primary (str): The column name for the primary column of the table.
        where_clause (str): The where clause that must be satisfied by the
            rows being refreshed.
        dry_run (bool): Set to True to execute a dry refresh run.
        avg_rows_per_second_cap (int): Number of rows we want to complete per second (sleeps in between batches to enforce)
    """
    notify_emails = ['bam+batch@yelp.com']
    is_readonly_batch = False
    DEFAULT_TOPOLOGY_PATH = "/nail/srv/configs/topology.yaml"
    DEFAULT_AVG_ROWS_PER_SECOND_CAP = 50

    def __init__(
        self,
        refresh_id=None,
        cluster=None,
        database=None,
        config_path=None,
        table_name=None,
        offset=None,
        batch_size=None,
        primary=None,
        where_clause=None,
        dry_run=False,
        avg_rows_per_second_cap=None
    ):
        super(FullRefreshRunner, self).__init__()
        self.config_path = config_path
        self._connection_set = None
        self.refresh_id = refresh_id
        # Case where the RefreshManager is running the refresh.
        if self.refresh_id is not None:
            self.topology_path = self.DEFAULT_TOPOLOGY_PATH
            signal.signal(signal.SIGTERM, self.handle_terminate)
            signal.signal(signal.SIGINT, self.handle_interupt)
            self.db_name = cluster
            self.database = database
            if not self.database:
                raise ValueError("--database must be specified")
            self.table_name = table_name
            self.temp_table = '{table}_data_pipeline_refresh'.format(
                table=self.table_name
            )
            self.processed_row_count = offset
            self.batch_size = batch_size
            if self.batch_size <= 0:
                raise ValueError("Batch size should be greater than 0")
            self.primary_key = primary
            self.where_clause = where_clause
            self.dry_run = dry_run
            self.avg_rows_per_second_cap = avg_rows_per_second_cap
            if self.avg_rows_per_second_cap is None:
                self.avg_rows_per_second_cap = self.DEFAULT_AVG_ROWS_PER_SECOND_CAP
            self.config_path = config_path

    @cached_property
    def schematizer(self):
        return get_schematizer()

    @property
    def version(self):
        """Overriding this so we'll get the clientlib version number when
        the tailer is run with --version.
        """
        return "data_pipeline {}".format(__version__)

    @batch_command_line_options
    def define_options(self, option_parser):
        opt_group = OptionGroup(option_parser, 'Full Refresh Runner Options')
        opt_group.add_option(
            '--cluster',
            dest='cluster',
            default='refresh_primary',
            help='Required: Specifies table cluster (default: %default).'
        )
        opt_group.add_option(
            '--database',
            dest='database',
            help='Specify the database to switch to after connecting to the '
                 'cluster.'
        )
        opt_group.add_option(
            '--table-name',
            dest='table_name',
            help='Required: Name of table to be refreshed.'
        )
        opt_group.add_option(
            '--batch-size',
            dest='batch_size',
            type='int',
            default=100,
            help='Number of rows to process between commits '
                 '(default: %default).'
        )
        opt_group.add_option(
            '--primary',
            dest='primary',
            help='Required: Comma separated string of primary key column names'
        )
        opt_group.add_option(
            '--dry-run',
            action="store_true",
            dest='dry_run',
            default=False
        )
        opt_group.add_option(
            '--config-path',
            dest='config_path',
            help='Required: Config file path for FullRefreshRunner '
                 '(default: %default)',
            default='/nail/srv/configs/data_pipeline_tools.yaml'
        )
        opt_group.add_option(
            '--topology-path',
            dest='topology_path',
            help='Path to the topology.yaml file.'
                 '(default: %default)',
            default=self.DEFAULT_TOPOLOGY_PATH
        )
        opt_group.add_option(
            '--where',
            dest='where_clause',
            default=None,
            help='Custom WHERE clause to specify which rows to refresh '
                 'Note: This option takes everything that would come '
                 'after the WHERE in a sql statement. '
                 'e.g: --where="country=\'CA\' AND city=\'Waterloo\'"'
        )
        opt_group.add_option(
            '--no-start-up-replication-wait',
            dest='wait_for_replication_on_startup',
            action='store_false',
            default=True,
            help='On startup, do not wait for replication to catch up to the '
                 'time batch was started (default: %default).'
        )
        opt_group.add_option(
            '--avg-rows-per-second-cap',
            help='Caps the throughput per second. Important since without any control for this '
                 'the batch can cause signifigant pipeline delays'
                 '(default: %default)',
            type='int',
            default=self.DEFAULT_AVG_ROWS_PER_SECOND_CAP
        )
        return opt_group

    @batch_configure
    def _init_global_state(self):
        if not self.config_path:
            self.config_path = self.options.config_path
        load_default_config(self.config_path)
        self._connection_set = None
        # Case where refresh batch is run independently.
        if self.refresh_id is None:
            if self.options.batch_size <= 0:
                raise ValueError("Batch size should be greater than 0")
            self.db_name = self.options.cluster
            self.database = self.options.database
            if not self.database:
                raise ValueError("--database must be specified")
            self.avg_rows_per_second_cap = self.options.avg_rows_per_second_cap
            if self.avg_rows_per_second_cap is not None and self.avg_rows_per_second_cap <= 0:
                raise ValueError("--avg-rows-per-second-cap should be greater than 0")
            self.table_name = self.options.table_name
            self.temp_table = '{table}_data_pipeline_refresh'.format(
                table=self.table_name
            )
            self.process_row_start_time = time.time()
            self.processed_row_count = 0
            self.batch_size = self.options.batch_size
            if self.batch_size <= 0:
                raise ValueError("Batch size should be greater than 0")
            self.primary_key = self.options.primary
            self.where_clause = self.options.where_clause
            self.dry_run = self.options.dry_run
            self.topology_path = self.options.topology_path

    def setup_connections(self):
        """Creates connections to the mySQL database.

        Builds a connection set to the cluster of the table that
         you are refreshing.

        This is overriding BatchDBMixin because we want to get connections
        based on the cluster instead of by replica names.
        TransactionManager also takes a custom connection_set_getter function
        which gets a connection set by cluster and topology file.
        """
        self._txn_mgr = TransactionManager(
            cluster_name=self.db_name,
            ro_replica_name=self.db_name,
            rw_replica_name=self.db_name,
            connection_set_getter=self.get_connection_set_from_cluster
        )

    @cached_property
    def total_row_count(self):
        with self.read_session() as session:
            self._use_db(session)
            query = self.build_select('COUNT(*)')
            value = self._execute_query(session, query).scalar()
            session.rollback()
            return value if value is not None else 0

    def build_select(
        self,
        select_item,
        order_col=None,
        offset=None,
        size=None
    ):
        base_query = 'SELECT {col} FROM {origin}'.format(
            col=select_item,
            origin=self.table_name
        )
        if self.where_clause is not None:
            base_query += ' WHERE {clause}'.format(clause=self.where_clause)
        if order_col is not None:
            base_query += ' ORDER BY {order}'.format(order=order_col)
        if offset is not None and size is not None:
            base_query += ' LIMIT {offset}, {size}'.format(
                offset=offset,
                size=size
            )
        return base_query

    def _wait_for_replication(self):
        """Lets first wait for ro_conn replication to catch up with the
        batch start time.
        """
        self.log.info(
            "Waiting for ro_conn replication to catch up with start time "
            "{start_time}".format(
                start_time=datetime.fromtimestamp(self.starttime)
            )
        )
        with self.ro_conn() as ro_conn:
            self.wait_for_replication_until(self.starttime, ro_conn)

    def create_table_from_src_table(self, session):
        show_create_statement = 'SHOW CREATE TABLE {table_name}'.format(
            table_name=self.table_name
        )
        original_query = self._execute_query(
            session,
            show_create_statement
        ).fetchone()[1]
        max_replacements = 1
        refresh_table_create_query = original_query.replace(
            self.table_name,
            self.temp_table,
            max_replacements
        )
        # Substitute original engine with Blackhole engine
        refresh_table_create_query = re.sub(
            'ENGINE=[^\s]*',
            'ENGINE=BLACKHOLE',
            refresh_table_create_query
        )
        refresh_table_create_query = re.sub(
            "CREATE TABLE",
            "CREATE TABLE IF NOT EXISTS",
            refresh_table_create_query,
            count=1
        )
        self.log.info("New blackhole table query: {query}".format(
            query=refresh_table_create_query
        ))
        self._execute_query(session, refresh_table_create_query)

    def _after_processing_rows(self, session, count):
        """Commits changes and makes sure replication and throughput cap
        catches up before moving on.
        """
        self._commit(session)

        # This code may seem counter-intuitive, but it's actually correct.
        # Tables shouldn't be unlocked until after the data changes have either
        # been committed or rolled back.  Committing or rolling back, does not,
        # however, unlock the locked tables.  The tables must be unlocked before
        # throttling to replication because failing to unlock them won't allow
        # replication to proceed, preventing replication from ever catching up.
        # session.commit() is called instead of _commit, since unlock tables
        # should be issued regardless of dry_run mode - it's necessary here
        # because sqlalchemy won't send the statement to MySQL without it.
        self._execute_query(session, 'UNLOCK TABLES')
        session.commit()

        with self.rw_conn() as rw_conn:
            self.throttle_to_replication(rw_conn)

        self._wait_for_throughput(count)

    def _wait_for_throughput(self, count):
        """Used to cap throughput when given the --avg-rows-per-second-cap flag.
        Sleeps for a certain amount of time based on elapsed time to process row, the number of rows processed (count)
        and the given cap so that the flag is enforced"""
        process_row_end_time = time.time()
        elapsed_time = process_row_end_time - self.process_row_start_time
        desired_elapsed_time = 1.0 / self.avg_rows_per_second_cap * count
        time_to_wait = max(desired_elapsed_time - elapsed_time, 0.0)
        self.log.info("Waiting for {} seconds to enforce avg throughput cap".format(time_to_wait))
        time.sleep(time_to_wait)

    def initial_action(self):
        self._wait_for_replication()
        with self.write_session() as session:
            if self.refresh_id:
                self.schematizer.update_refresh(
                    self.refresh_id,
                    RefreshStatus.IN_PROGRESS,
                    self.processed_row_count
                )
            self._use_db(session)
            self.create_table_from_src_table(session)
            self._commit(session)

    def _use_db(self, session):
        self._execute_query(
            session,
            "USE {database}".format(database=self.database)
        )

    def _commit(self, session):
        """Commits unless in dry_run mode, otherwise rolls back"""
        if self.dry_run:
            self.log.info("Executing rollback in dry-run mode")
            session.rollback()
        else:
            session.commit()

    def final_action(self):
        with self.write_session() as session:
            self._use_db(session)
            query = 'DROP TABLE IF EXISTS {temp_table}'.format(
                temp_table=self.temp_table
            )
            self._execute_query(session, query)
            self.log.info("Dropped table: {table}".format(table=self.temp_table))
            self._commit(session)

    def setup_transaction(self, session):
        self._use_db(session)
        self._execute_query(
            session,
            'LOCK TABLES {table} WRITE, {temp} WRITE'.format(
                table=self.table_name,
                temp=self.temp_table
            )
        )

    def count_inserted(self, session, min_pk, max_pk):
        select_query = self._get_select_query(session, min_pk, max_pk)
        query = 'SELECT COUNT(*) FROM ({query}) AS T'.format(
            query=select_query
        )
        inserted_rows = self._execute_query(session, query)
        return inserted_rows.scalar()

    def insert_batch(self, session, min_pk, max_pk):
        insert_query = 'INSERT INTO {temp}'.format(temp=self.temp_table)
        select_query = self._get_select_query(session, min_pk, max_pk)
        insert_query += select_query
        self._execute_query(session, insert_query)

    def _get_select_query(self, session, min_pk, max_pk):
        select_query = """
        SELECT * FROM {table}
        WHERE {pk}>{min_pk} AND {pk}<={max_pk}
        """.format(
            table=self.table_name,
            pk=self.primary_key,
            min_pk=min_pk,
            max_pk=max_pk
        )
        if self.where_clause is not None:
            select_query += ' AND {clause}'.format(clause=self.where_clause)

        return select_query

    def _get_min_primary_key(self):
        select_query = """
        SELECT MIN({pk}) FROM {table}
        """.format(
            pk=self.primary_key,
            table=self.table_name
        )
        with self.read_session() as session:
            min_pk = self._execute_query(session, select_query)
            session.commit()
        return min_pk.scalar()

    def _get_max_primary_key(self):
        select_query = """
        SELECT MAX({pk}) FROM {table}
        """.format(
            pk=self.primary_key,
            table=self.table_name
        )

        with self.read_session() as session:
            max_pk = self._execute_query(session, select_query)
            session.commit()
        return max_pk.scalar()

    def process_table(self):
        self.log.info(
            "Total rows to be processed: {row_count}".format(
                row_count=self.total_row_count
            )
        )
        min_pk = self._get_min_primary_key() - 1
        max_pk = self._get_max_primary_key()
        while min_pk < max_pk:
            self.process_row_start_time = time.time()
            with self.write_session() as session:
                self.setup_transaction(session)
                min_batch = min_pk + self.batch_size
                inserted = self.count_inserted(session, min_pk, min_batch)
                self.insert_batch(session, min_pk, min_batch)
                self._after_processing_rows(session, inserted)
            min_pk = min_batch
            self.processed_row_count += inserted

        if self.refresh_id:
            # Offset is 0 because it doesn't matter (was a success)
            self.schematizer.update_refresh(refresh_id=self.refresh_id, status=RefreshStatus.SUCCESS, offset=0)

    def log_info(self):
        elapsed_time = time.time() - self.starttime
        self.log.info(
            "Processed {row_count} row(s) in {elapsed_time}".format(
                row_count=self.processed_row_count,
                elapsed_time=elapsed_time
            )
        )

    def handle_interupt(self, signal, frame):
        self.schematizer.update_refresh(
            self.refresh_id,
            RefreshStatus.FAILED,
            self.processed_row_count
        )
        os._exit(1)

    def handle_terminate(self, signal, frame):
        self.schematizer.update_refresh(
            self.refresh_id,
            RefreshStatus.PAUSED,
            self.processed_row_count
        )
        os._exit(1)

    def run(self):
        is_success = True
        try:
            self.initial_action()
            self.process_table()
            self.log_info()
        except SystemExit:
            pass
        except Exception:
            if self.refresh_id:
                self.schematizer.update_refresh(
                    self.refresh_id,
                    RefreshStatus.FAILED,
                    self.processed_row_count
                )
            # Sends an email containing the exception encountered.
            self._email_exception_in_exception_context()
            is_success = False
        finally:
            # We don't want to drop the blackhole table until the replication handler is using
            # the schema_tracker database stably. (TODO: DATAPIPE-845)
            # self.final_action()
            if not is_success:
                os._exit(1)

    def get_connection_set_from_cluster(self, cluster):
        """Given a cluster name, returns a connection to that cluster.
        """
        if self._connection_set:
            return self._connection_set
        topology = TopologyFile.new_from_file(self.topology_path)
        conn_defs = self._get_conn_defs(topology, cluster)
        conn_config = ConnectionSetConfig(cluster, conn_defs, read_only=False)
        self._connection_set = ConnectionSet.from_config(conn_config)
        return self._connection_set

    def _get_conn_defs(self, topology, cluster):
        replica_level = 'master'
        connection_cluster = topology.topologies[cluster, replica_level]
        conn_def = ConnectionDef(
            cluster,
            replica_level,
            auto_commit=False,
            database=connection_cluster.database
        )
        return {cluster: (conn_def, connection_cluster)}

    def _execute_query(self, session, query):
        self.log.debug("Executing query: {query}".format(query=query))
        return session.execute(query)


if __name__ == '__main__':
    FullRefreshRunner().start()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from enum import Enum


class ExpectedFrequency(Enum):
    """Helper constants specifying how frequently the client expects to produce
    or consume messages.  Any positive integer number of seconds can be used,
    these are provided for convenience only.  Expected frequency will be used
    to infer schema deprecation.

    For example, if a client registers to produce messages constantly, and a
    few months later we observe that the client hasn't published messages using
    an older schema verison in a month, but does regularly publish using a
    newer version, we can infer that the older schema version is deprecated
    and send out a deprecation/migration notice.

    Attributes:
      constantly: client expects to always and continuously be producing and
        consuming messages.
      hourly: client expects to come online to produce or consume messages
        approximately every hour.
      weekly: client expects to come online to produce or consume messages about
        once a week.
      monthly: client expects to come online to produce or consume messages about
        once a month.
      yearly: client expects to come online to produce or consume messages about
        once a year.
    """
    constantly = 0
    hourly = 60 * 60
    daily = hourly * 24
    weekly = daily * 7
    monthly = daily * 30
    yearly = daily * 365



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import errno
from time import time

from kafka.common import ConsumerTimeout
from yelp_kafka.consumer_group import KafkaConsumerGroup

from data_pipeline.base_consumer import BaseConsumer
from data_pipeline.config import get_config
from data_pipeline.message import create_from_kafka_message

logger = get_config().logger


class Consumer(BaseConsumer):
    """
    The Consumer uses an iterator to get messages that need to be consumed
    from Kafka.

    Args:
        consumer_name (str): See parameter `client_name` in
            :class:`data_pipeline.client.Client`.  The `consumer_name` will
            be registered with Kafka to commit offsets.
        team_name (str): See parameter `team_name` in
            :class:`data_pipeline.client.Client`.
        expected_frequency_seconds (int, ExpectedFrequency): See parameter
            `expected_frequency_seconds` in :class:`data_pipeline.client.Client`.
        topic_to_consumer_topic_state_map ({str:Optional(ConsumerTopicState)}):
            A map of topic names to `ConsumerTopicState` objects which define
            the offsets to start from. The ConsumerTopicState of a topic may be
            `None`, in which case the committed kafka offset for the
            consumer_name is used. If there is no committed kafka offset for
            the consumer_name the consumer will begin from the
            `auto_offset_reset` offset in the topic.
        consumer_source (ConsumerSource): Object to specify the topics this
            consumer consumes messages from. It must be a
            :class:`data_pipeline.consumer_source.ConsumerSource` object. For
            example, to process messages from a fixed set of topics, use
            :class:`data_pipeline.consumer_source.FixedTopics`.
            In case of FixedSchema consumer source, at most one schema_id can
            be provided per topic. Consumer would use that schema as reader
            schema to decode the message. In case no schema id for a topic is
            specified, then the Consumer would use the schema id that the
            message was encoded with to decode the message.
        auto_offset_reset (str): automatically resets the offset when there is
            no initial offset in Zookeeper or if an offset is out of range.
            If 'largest', reset the offset to the latest available message (tail).
            If 'smallest' reset from the earliest (head).
        partitioner_cooldown (float): Waiting time (in seconds) for the
            consumer to acquire the partitions. See
            yelp_kafka/yelp_kafka/partitioner.py for more details
        use_group_sha (Optional[boolean]): Used by partitioner to establish
            group membership. If false, consumer group with same name will
            be treated as the same group; otherwise, they will be different
            since group sha is different. Default is true.
        pre_rebalance_callback (Optional[Callable[{str:list[int]}, None]]):
            Optional callback which is passed a dict of topic as key and list
            of partitions as value. It's important to note this may be called
            multiple times in a single repartition, so any actions taken as a
            result must be idempotent. You are guaranteed that no messages will
            be consumed between this callback and the post_rebalance_callback.
        post_rebalance_callback (Optional[Callable[{str:list[int]}, None]]):
            Optional callback which is passed a dict of topic as key and list
            of partitions as value which were acquired in a repartition. You
            are guaranteed that no messages will be consumed between the
            pre_rebalance_callback and this callback.
        fetch_offsets_for_topics: (Optional[Callable[List[str],
            Dict[str, Optional[Dict[int, int]]]]]): Optional callback which is
            passed a list of topics, and should return a dictionary where keys
            are topic names and values are either None if no offset should be
            manually set, or a map from partition to offset.
            If implemented, this function will be called every time
            consumer refreshes the topics, so that the consumer can provide a
            map of partitions to offsets for each topic, or None if the default
            behavior should be employed instead. The default behavior is
            picking up the last committed offsets of topics.
            This method must be implemented if topic state is to be stored
            in some system other than Kafka, for example when writing data from
            Kafka into a transactional store.
        pre_topic_refresh_callback: (Optional[Callable[[set[str], set[str]],
            Any]]): Optional callback that gets executed right before the
            consumer is about to refresh the topics. The callback function is
            passed in a set of topic names Consumer is currently consuming
            from (current_topics) and a set of topic names Consumer will be
            consuming from (refreshed_topics). The return value of the
            function is ignored.

    Note:
        The Consumer leverages the yelp_kafka `KafkaConsumerGroup`.

    **Examples**:

    A simple example can be a consumer with name 'my_consumer' that
    consumes a message from multiple topics, processes it and
    commits the offset and this process continues::

        with Consumer(
            consumer_name='my_consumer',
            team_name='bam',
            expected_frequency_seconds=12345,
            topic_to_consumer_topic_state_map={
                'topic_a': None,
                'topic_b': None
            }
        ) as consumer:
            while True:
                message = consumer.get_message()
                if message is not None:
                    ... do stuff with message ...
                    consumer.commit_message(message)

    Note:
        Recommended to avoid calling `commit_message(message)` after every
        message, as it is relatively expensive.

    Another example can be a consumer which consumes multiple messages
    (with maximum number of messages in batch as `count`) from 2 topics
    'topic_a' and 'topic_b', processes them and commits them::

        with Consumer(
            consumer_name='my_consumer',
            team_name='bam',
            expected_frequency_seconds=12345,
            topic_to_consumer_topic_state_map={
                'topic_a': None,
                'topic_b': None
            }
        ) as consumer:
            while True:
                messages = consumer.get_messages(
                    count=batch_size,
                    blocking=True,
                    timeout=batch_timeout
                )
                if messages:
                    ... do stuff with messages ...
                    consumer.commit_messages(messages)

    Note:
        It's recommended to retrieve messages in batches via
        `get_messages(..)`, do your work with them, and then commit them as
        a group with a single call to `commit_messages(..)`
    """

    def _start(self):
        self.consumer_group = KafkaConsumerGroup(
            topics=self.topic_to_partition_map.keys(),
            config=self._kafka_consumer_config
        )
        self.consumer_group.start()

    def _stop(self):
        self.consumer_group.stop()

    def get_messages(
            self,
            count,
            blocking=False,
            timeout=get_config().consumer_get_messages_timeout_default
    ):
        """ Retrieve a list of messages from the message buffer, optionally
        blocking until the requested number of messages has been retrieved.

        Warning:
            If `blocking` is True and `timeout` is None this will block until
            the requested number of messages is retrieved, potentially blocking
            forever. Please be absolutely sure this is what you are intending
            if you use these options!

        Args:
            count (int): Number of messages to retrieve
            blocking (boolean): Set to True to block while waiting for messages
                if the buffer has been depleted. Otherwise returns immediately
                if the buffer reaches depletion. Default is False.
            timeout (double): Maximum time (in seconds) to wait if blocking is
                set to True. Set to None to wait indefinitely.

        Returns:
            ([data_pipeline.message.Message]): List of Message objects with
            maximum size `count`, but may be smaller or empty depending on
            how many messages were retrieved within the timeout.
        """
        # TODO(tajinder|DATAPIPE-1231): Consumer should refresh topics
        # periodically even if NO timeout is provided and there are no
        # messages to consume.
        messages = []
        has_timeout = timeout is not None
        if has_timeout:
            max_time = time() + timeout
        while len(messages) < count:
            # Consumer refreshes the topics periodically only if consumer_source
            # is specified and would use the `fetch_offsets_for_topics` callback
            # to get the partition offsets corresponding to the topics.
            if self.consumer_source:
                self._refresh_source_topics_if_necessary()
            try:
                default_iter_timeout = self.consumer_group.iter_timeout
                # Converting seconds to milliseconds
                self.consumer_group.iter_timeout = timeout * 1000
                kafka_message = self._get_next_kafka_message(
                    blocking,
                    has_timeout,
                    max_time
                )

                # It's possible kafka_message is None if we used all our time
                # stuck getting EINTR IOErrors
                if kafka_message:
                    message = create_from_kafka_message(
                        kafka_message,
                        self._envelope,
                        self.force_payload_decode,
                        reader_schema_id=self._topic_to_reader_schema_map.get(
                            kafka_message.topic
                        )
                    )
                    messages.append(message)
                    # Update state in registrar for Producer/Consumer
                    # registration in milliseconds
                    self.registrar.update_schema_last_used_timestamp(
                        message.reader_schema_id,
                        timestamp_in_milliseconds=long(1000 * time())
                    )
                if self._break_consume_loop(blocking, has_timeout, max_time):
                    break
            except ConsumerTimeout:
                break
            finally:
                self.consumer_group.iter_timeout = default_iter_timeout
        return messages

    def _get_next_kafka_message(
            self,
            blocking,
            has_timeout,
            max_time
    ):
        """ Helper function which will retry when encountering an IOError with
        the errno of EINTR. This is now standard behavior in Python3.5. For
        more details see https://www.python.org/dev/peps/pep-0475/
        """
        while not self._break_consume_loop(blocking, has_timeout, max_time):
            try:
                return self.consumer_group.next()
            except IOError as e:
                if e.errno != errno.EINTR:
                    raise
        return None

    def _break_consume_loop(self, blocking, has_timeout, max_time):
        return not blocking or (has_timeout and time() > max_time)

    def _refresh_source_topics_if_necessary(self):
        if not self._refresh_timer.should_tick():
            return

        current_topics = set(self.topic_to_partition_map.keys())
        refreshed_topics = set(self.consumer_source.get_topics())

        if current_topics == refreshed_topics:
            return

        all_topics_to_state_map = self._get_topic_to_offset_map(
            current_topics.union(refreshed_topics)
        )
        refreshed_topics_to_state_map = {
            topic: all_topics_to_state_map.get(topic)
            for topic in refreshed_topics
        }

        if self.pre_topic_refresh_callback:
            self.pre_topic_refresh_callback(current_topics, refreshed_topics)

        self.stop()
        self._commit_topic_offsets(all_topics_to_state_map)
        self._set_topic_to_partition_map(refreshed_topics_to_state_map)
        self._start_consumer()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple


class PositionData(namedtuple("PositionData", [
    "last_published_message_position_info",
    "topic_to_last_position_info_map",
    "topic_to_kafka_offset_map",
    "merged_upstream_position_info_map"
])):
    """Contains information about the last messages successfully published into
    Kafka.

    **Examples**:

        When first activated, position data is minimal:

        >>> from data_pipeline.producer import Producer
        >>> with Producer() as producer:
        ...     position_data = producer.get_checkpoint_position_data()
        >>> position_data.last_published_message_position_info == None
        True
        >>> position_data.topic_to_last_position_info_map
        {}
        >>> position_data.topic_to_kafka_offset_map
        {}

        :class:`PositionData` will be updated as data is actually published:

        >>> from data_pipeline.message import CreateMessage
        >>> with Producer() as producer:
        ...     producer.publish(CreateMessage(
        ...         schema_id=10,
        ...         payload=bytes(10),
        ...         upstream_position_info={str('upstream_offset'): str('offset-info')}
        ...     ))
        ...     producer.flush()
        ...     position_data = producer.get_checkpoint_position_data()
        >>> position_data.last_published_message_position_info
        {'upstream_offset': 'offset-info'}
        >>> position_data.topic_to_last_position_info_map
        {'my-topic': {'upstream_offset': 'offset-info'}}
        >>> position_data.topic_to_kafka_offset_map   # doctest: +ELLIPSIS
        {'my-topic': ...}
        >>> position_data.merged_upstream_position_info_map
        {'upstream_offset': 'offset-info'}

    Warning:

        :class:`PositionData` returned by
        :meth:`data_pipeline.producer.Producer.get_checkpoint_position_data` is
        only the source of truth for topics that have messages published to
        them during the current session.  If this data is persisted, it should
        be merged with data from prior sessions, preferring data from the
        current session only when it is available.

    Attributes:
        last_published_message_position_info (None, dict): If a message has been
            published successfully, this will contain the
            :attr:`data_pipeline.message.Message.upstream_position_info` from
            the last message that was successfully published, irrespective of
            topic.  In this context, last refers to messages sorted by the
            order they were passed into :meth:`Producer.publish`.  This field
            is meant to be used if upstream data originates from a single
            source.

            **Example**:

                Assuming a dictionary containing gtid and offset information is
                passed into
                :attr:`data_pipeline.message.Message.upstream_position_info`::

                    {'gtid': 'UPSTREAM_GTID_INFO', 'offset': 'UPSTREAM_OFFSET_INFO'}

        topic_to_last_position_info_map (dict[str, dict]): Maps from topic to
            the last :attr:`data_pipeline.message.Message.upstream_position_info`
            that was successfully published for each topic, where last refers to
            messages sorted by the order they were passed into
            :meth:`Producer.publish`.  The dictionary starts empty, and only
            contains information about topics published into during this
            session.  This field is meant to be used if upstream data
            originates from multiple sources.

            **Example**:

                The dictionary will contain the last upstream_position_info
                for each topic.  Assuming a message had only been published to
                `my-topic`::

                    {'my-topic': {'gtid': 'UPSTREAM_GTID_INFO', 'offset': 'UPSTREAM_OFFSET_INFO'}}

        topic_to_kafka_offset_map (dict[str, int]): This maps from each kafka
            topic to the offset following the last published message, which will
            correspond to the offset of the next message published into the
            topic.

            **Example**:

                The dictionary will contain a Kafka offset for each topic
                publised to in this session::

                    {'topic1': offset1, 'topic2': offset2}

        merged_upstream_position_info_map (dict): This dictionary starts empty,
            and contains a deep merge of
            :attr:`data_pipeline.message.Message.upstream_position_info`
            for all messages that have been successfully published.  This
            structure is particularly useful to track the combined upstream
            position info for multiple topics.

            **Example**:

                This dictionary starts empty, and each messages
                `upstream_position_info` will be deep-merged into it::

                    {'topic1': {0: offset1}, 'topic2': {0, offset2}}

                When implementing a kafka-to-kafka transformer pipeline, this
                makes it easy to build up a structure to pass into
                :meth:`data_pipeline.base_consumer.BaseConsumer.commit_offsets`.
                Each transformed message would include `upstream_position_info`
                like::

                    {topic: {partition: offset}}
    """
    # This is a class instead of a namedtuple so the docstring can be
    # set.



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Use this like:
mysqlbinlog --read-from-remote-server --host 10.69.1.100 -u rbr_test \
    --stop-never --verbose --start-datetime="2015-03-08 00:45:00" \
    mysql-bin.000405 | ~/pypy-2.5.0-linux64/bin/pypy ~/binlog_analyzer.py \
    | ~/pypy-2.5.0-linux64/bin/pypy ~/compressed_stream_rotator.py
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import errno
import fileinput
import json
import re
import time


class BinlogParser(object):
    statement_to_type = {'INSERT INTO': 'insert', 'UPDATE': 'update', 'DELETE FROM': 'delete'}

    def __init__(self):
        self.timestamp = None
        self.header_timestamp = None

    def run(self):
        try:
            self._parse_binlog()
        except IOError as e:
            if e.errno == errno.EPIPE:
                # just stop if the pipe breaks
                pass
            else:
                raise

    def _parse_binlog(self):
        for line in fileinput.input():
            line = line.strip()
            self._process_line(line)

    def _process_line(self, line):
        if self._is_setting_timestamp(line):
            self._handle_timestamp_line(line)
        if self._is_header_line(line):
            self._handle_header_line(line)
        elif self._is_updating(line):
            self._handle_update_line(line)

    def _is_setting_timestamp(self, line):
        return line.startswith("SET TIMESTAMP=") and line.endswith("/*!*/;")

    def _handle_header_line(self, line):
        m = re.search("\\#(\\d+)\\s+(\\d+:\\d+:\\d+)\\s+server\\s+id\\s+\\d+", line)
        datetime_str = "%s %s" % (m.group(1), m.group(2))
        dt = datetime.datetime.strptime(datetime_str, '%y%m%d %H:%M:%S')
        new_header_timestamp = int(time.mktime(dt.timetuple()))
        self.header_timestamp = new_header_timestamp

    def _is_header_line(self, line):
        regex = "\\#(\\d+)\\s+(\\d+:\\d+:\\d+)\\s+server\\s+id\\s+\\d+.+(Update_rows|Write_rows|Delete_rows)"
        return re.search(regex, line) is not None

    def _handle_timestamp_line(self, line):
        m = re.search("SET\\ TIMESTAMP=(\\d+)/\\*!\\*/;", line)
        new_timestamp = int(m.group(1))
        self.timestamp = new_timestamp

    def _is_updating(self, line):
        return any(line.startswith("### %s " % s) for s in ['INSERT INTO', 'UPDATE', 'DELETE FROM'])

    def _handle_update_line(self, line):
        m = re.search("\\#\\#\\#\\ (DELETE\\ FROM|INSERT\\ INTO|UPDATE)\\ (.+)", line)
        statement_type = self.statement_to_type[m.group(1)]
        table = m.group(2)

        print json.dumps({
            'timestamp': self.header_timestamp,
            'statement_type': statement_type,
            'table': table
        })


if __name__ == "__main__":
    BinlogParser().run()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import warnings
from collections import namedtuple

import simplejson
from data_pipeline_avro_util.util import get_avro_schema_object

from data_pipeline.config import get_config


# TODO([DATAPIPE-396|clin]): Revise these namedtuples to their own classes
# as soon as we can! They should be separate classes in their own files in
# separate folder along with this schematizer clientlib module.
SchemaInfo = namedtuple('SchemaInfo', [
    'schema_id',
    'topic_name'
])

AvroSchema = namedtuple(
    'AvroSchema',
    ['schema_id', 'schema', 'topic', 'base_schema_id', 'created_at']
)
Topic = namedtuple('Topic', ['topic_id', 'name', 'source', 'created_at'])
Source = namedtuple('Source', ['source_id', 'name', 'namespace', 'created_at'])
Namespace = namedtuple('Namespace', ['namespace_id', 'name', 'created_at'])


class SchematizerClient(object):
    """This old Schematizer cache is now deprecated.

    Client that provides high level api calls to interact with Schematizer
    service.  It has a built-in cache that maps schema_id's to their schemas
    and to their transformed schema_ids.

        Currently this only holds an in-memory cache.
        TODO(DATAPIPE-162|joshszep): Implement persistent caching
    """

    def __init__(self):
        warnings.simplefilter('always', category=DeprecationWarning)
        warnings.warn("data_pipeline.schema_cache.SchematizerClient is deprecated.",
                      DeprecationWarning)
        self.schema_id_to_schema_map = {}
        self.schema_id_to_topic_map = {}
        self.base_to_transformed_schema_id_map = {}
        self.schema_id_to_pii_map = {}

    @property
    def schematizer_client(self):
        """TODO[DATAPIPE-396|clin]: change this to be private once this class
        is converted to the true schematizer client.
        """
        return get_config().schematizer_client

    def get_transformed_schema_id(self, schema_id):
        """ Get the cached transformed schema_id corresponding to the given
            base_schema_id if it is known, otherwise None is returned

        Args:
            schema_id (int): The base schema_id

        Returns:
            int: cached transformed schema_id if known, None otherwise
        """
        return self.base_to_transformed_schema_id_map.get(schema_id, None)

    def _register_schema(
        self,
        namespace,
        source,
        schema_str,
        owner_email,
        contains_pii,
        base_schema_id=None
    ):
        """This is an initial step to rename the existing `register_transformed_schema`
        function. The function is to register an Avro schema, with optional base
        schema. So keep the name neutral and not AST specific. For new code, please
        use this function. The existing function `register_transformed_schema` will
        continue to work until we're safe to rename it.

        Right now the function is private and will stay private until it's ready.
        """
        request_body = {
            'schema': schema_str,
            'namespace': namespace,
            'source': source,
            'source_owner_email': owner_email,
            'contains_pii': contains_pii,
        }
        if base_schema_id:
            request_body['base_schema_id'] = base_schema_id
        schema_resp = self.schematizer_client.schemas.register_schema(
            body=request_body
        ).result()
        # TODO[DATAPIPE-396|clin]: add caching as part of DATAPIPE-396
        return self._construct_schema(schema_resp)

    def register_schema_by_schema_json(
        self,
        namespace,
        source,
        schema_json,
        owner_email,
        contains_pii,
        base_schema_id=None
    ):
        """Register a new avro schema.

        Args:
            namespace (str): The namespace the new schema belongs to
            source (str): The source the new schema belongs to.
            schema_json (dict or list): The Python object representation of the
                new avro schema json.
            owner_email (str): The email of the source owner.
            contains_pii (bool): Indicates whether the new schema contains
                field(s) that store pii information.  See http://y/pii for
                help identifying what is or is not PII.
            base_schema_id (Optional[int]): The schema_id of the original schema
                which the new schema is changed based on.

        Returns:
            `data_pipeline.schema_cache.AvroSchema`: The object containing the
                information of newly created avro schema.
        """
        return self._register_schema(
            namespace=namespace,
            source=source,
            schema_str=simplejson.dumps(schema_json),
            owner_email=owner_email,
            contains_pii=contains_pii,
            base_schema_id=base_schema_id
        )

    def register_transformed_schema(
        self,
        base_schema_id,
        namespace,
        source,
        schema,
        owner_email,
        contains_pii
    ):
        """ Register a new schema and return it's schema_id and topic.
        Note that this function is going to be deprecated soon.  Please use
        `register_schema` function instead for moving forward.

        Args:
            base_schema_id (int): The schema_id of the original schema from
                which the the new schema was transformed
            namespace (str): The namespace the new schema should be registered
                to.
            source (str): The source the new schema should be registered to.
            schema (str): The new schema in json string representation.
            owner_email (str): The owner email for the new schema.
            contains_pii (bool): Indicates that the schema being registered has
                at least one field that can potentially contain PII.
                See http://y/pii for help identifying what is or is not PII.

        Returns:
            (int, string): The new schema_id and the new topic name
        """
        request_body = {
            'base_schema_id': base_schema_id,
            'schema': schema,
            'namespace': namespace,
            'source': source,
            'source_owner_email': owner_email,
            'contains_pii': contains_pii,
        }
        register_response = self.schematizer_client.schemas.register_schema(
            body=request_body
        ).result()
        transformed_id = register_response.schema_id
        self.schema_id_to_schema_map[transformed_id] = register_response.schema
        self.base_to_transformed_schema_id_map[base_schema_id] = transformed_id
        new_topic_name = register_response.topic.name
        self.schema_id_to_topic_map[transformed_id] = new_topic_name
        return SchemaInfo(schema_id=register_response.schema_id, topic_name=new_topic_name)

    def register_schema_from_mysql_stmts(
            self,
            new_create_table_stmt,
            namespace,
            source,
            owner_email,
            contains_pii,
            old_create_table_stmt=None,
            alter_table_stmt=None,
    ):
        """ Register schema based on mysql statements and return it's schema_id
            and topic.

        Args:
            new_create_table_stmt (str): the mysql statement of creating new table.
            namespace (str): The namespace the new schema should be registered to.
            source (str): The source the new schema should be registered to.
            owner_email (str): The owner email for the new schema.
            contains_pii (bool): The flag indicating if schema contains pii.
            old_create_table_stmt (str optional): the mysql statement of creating old table.
            alter_table_stmt (str optional): the mysql statement of altering table schema.

        Returns:
            (int, string): The new schema_id and the new topic name
        """
        request_body = {
            'new_create_table_stmt': new_create_table_stmt,
            'namespace': namespace,
            'source': source,
            'source_owner_email': owner_email,
            'contains_pii': contains_pii
        }
        if old_create_table_stmt:
            request_body['old_create_table_stmt'] = old_create_table_stmt
        if alter_table_stmt:
            request_body['alter_table_stmt'] = alter_table_stmt
        register_response = self.schematizer_client.schemas.register_schema_from_mysql_stmts(
            body=request_body
        ).result()
        schema_id = register_response.schema_id
        self.schema_id_to_schema_map[schema_id] = register_response.schema
        new_topic_name = register_response.topic.name
        self.schema_id_to_topic_map[schema_id] = new_topic_name
        return SchemaInfo(schema_id=register_response.schema_id, topic_name=new_topic_name)

    def get_topic_for_schema_id(self, schema_id):
        """ Get the topic name for a given schema_id

        Args:
            schema_id (int): The schema_id you are curious about.

        Returns:
            (str): The topic name for the given schema_id
        """
        if schema_id not in self.schema_id_to_topic_map:
            self.schema_id_to_topic_map[schema_id] = self._retrieve_topic_name_from_schematizer(schema_id)
        return self.schema_id_to_topic_map[schema_id]

    def get_contains_pii_for_schema_id(self, schema_id):
        if schema_id not in self.schema_id_to_pii_map:
            self.schema_id_to_pii_map[schema_id] = self._retrieve_contains_pii_from_schematizer(schema_id)
        return self.schema_id_to_pii_map[schema_id]

    def get_schema(self, schema_id):
        """ Get the schema corresponding to the given schema_id, handling cache
            misses if required

        Args:
            schema_id (int): The schema_id to use for lookup.

        Returns:
            (avro.schema.Schema): The avro Schema object
        """
        if schema_id not in self.schema_id_to_schema_map:
            self.schema_id_to_schema_map[schema_id] = self._retrieve_avro_schema_from_schematizer(schema_id)
        return self.schema_id_to_schema_map[schema_id]

    def _get_schema_from_schematizer(self, schema_id):
        # TODO(DATAPIPE-207|joshszep): Include retry strategy support
        return self.schematizer_client.schemas.get_schema_by_id(
            schema_id=schema_id
        ).result()

    def _retrieve_topic_name_from_schematizer(self, schema_id):
        return self._get_schema_from_schematizer(schema_id).topic.name

    def _retrieve_contains_pii_from_schematizer(self, schema_id):
        return self._get_schema_from_schematizer(schema_id).topic.contains_pii

    def _retrieve_avro_schema_from_schematizer(self, schema_id):
        return get_avro_schema_object(
            self._get_schema_from_schematizer(schema_id).schema
        )

    def get_topics_by_criteria(
        self,
        namespace_name=None,
        source_name=None,
        created_after=None
    ):
        """Get all the topics that match specified criteria.  If no criterion
        is specified, it returns all the topics.

        Args:
            namespace_name (Optional[str]): namespace the topics belong to
            source_name (Optional[str]): name of the source topics belong to
            created_after (Optional[int]): Epoch timestamp the topics should be
                created after.  The topics created at the same timestamp are
                also included.

        Returns:
            [schema_cache.Topic]: list of topics that match given criteria.
        """
        topics_resp = self.schematizer_client.topics.get_topics_by_criteria(
            namespace=namespace_name,
            source=source_name,
            created_after=created_after
        ).result()
        return [self._construct_topic(t) for t in topics_resp]

    def _construct_schema(self, response):
        # TODO([DATAPIPE-396|clin]): This should be replaced with class constructor
        # once the AvroSchema namedtuple is replaced with the class.
        return AvroSchema(
            schema_id=response.schema_id,
            schema=response.schema,
            topic=self._construct_topic(response.topic),
            base_schema_id=response.base_schema_id,
            created_at=response.created_at
        )

    def _construct_topic(self, response):
        # TODO([DATAPIPE-396|clin]): This should be replaced with class constructor
        # once the Topic namedtuple is replaced with the class.
        return Topic(
            topic_id=response.topic_id,
            name=response.name,
            source=self._construct_source(response.source),
            created_at=response.created_at
        )

    def _construct_source(self, response):
        # TODO([DATAPIPE-396|clin]): This should be replaced with class constructor
        # once the Source namedtuple is replaced with the class.
        return Source(
            source_id=response.source_id,
            name=response.name,
            namespace=self._construct_namespace(response.namespace),
            created_at=response.created_at
        )

    def _construct_namespace(self, response):
        # TODO([DATAPIPE-396|clin]): This should be replaced with class constructor
        # once the Namespace namedtuple is replaced with the class.
        return Namespace(
            namespace_id=response.namespace_id,
            name=response.name,
            created_at=response.created_at
        )


_schematizer_client = SchematizerClient()


def get_schema_cache():
    return _schematizer_client


def get_schematizer_client():
    """This is an initial step for DATAPIPE-396 to eventually rename the
    schema cache to schematizer client.  For new code that needs to access
    schema_cache, use this function instead of `get_schema_cache` function.
    For the existing code, `get_schema_cache` will continue to work in the
    meantime.
    """
    return _schematizer_client



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline._avro_payload import _AvroPayload


class MetaAttribute(object):
    """Messages flowing through data pipeline can contain an
    additional array of avro encoded payloads under the “meta” key.
    These avro encoded payloads are known as Meta Attributes within the
    data pipeline domain. Meta Attributes may contains additional information
    alongside messages originating from data pipeline producer.

    For example for messages coming from metrics source within yelp_web_metric
    namespace, we could include the web request itself as one of the
    Meta Attribute in that message.

    1. Get a new meta attribute object from `schema_id` and `payload_data`

    Example:

        MetaAttribute(
            schema_id=schema_id,
            payload_data={'cluster_name': 'cluster1',
            'log_file': 'binlog.0001', 'log_pos': 12332})

    2. Recover a MetaAttribute object from serialized MetaAttribute payload.
    This will be usefull specifically in the case where we want to reconstruct
    MetaAttribute object from serialized payload coming out of Kafka.

    Example:

        MetaAttribute(schema_id=schema_id,payload=byte(10))

    Args:
        schema_id (int): Identifies the schema used to encode the payload.
        payload (bytes): Avro-encoded meta attribute - encoded with schema identified
            by `schema_id`. Either `payload` or `payload_data` must be
            provided but not both.
        payload_data: The contents of meta attribute, which will be
            encoded with schema identified by `schema_id`. Either `payload` or
            `payload_data` must be provided but not both. Type of payload_data
            should match the avro type specified schema.
        dry_run (boolean): When set to True, MetaAttribute will return a string
            representation of the payload, instead of the avro encoded MetaAttribute.
            Defaults to False.
    """

    def __init__(
        self,
        schema_id,
        payload=None,
        payload_data=None,
        dry_run=False
    ):
        self._avro_payload = _AvroPayload(
            schema_id=schema_id,
            payload=payload,
            payload_data=payload_data,
            dry_run=dry_run
        )

    @property
    def payload(self):
        return self._avro_payload.payload

    @property
    def payload_data(self):
        return self._avro_payload.payload_data

    @property
    def schema_id(self):
        return self._avro_payload.schema_id

    @property
    def avro_repr(self):
        return {
            'schema_id': self.schema_id,
            'payload': self.payload
        }

    def _asdict(self):
        """ Helper method for simplejson encoding in the Tailer
        """
        return {
            'schema_id': self.schema_id,
            'payload_data': self._avro_payload.printable_payload_data
        }

    def __repr__(self):
        return '{}'.format(self._asdict())



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.tools.introspector.models.base import BaseIntrospectorModel


class IntrospectorTopic(BaseIntrospectorModel):
    def __init__(self, topic_obj, kafka_topics=None, topics_to_range_map=None):
        super(IntrospectorTopic, self).__init__(
            topic_obj,
            excluded_fields=['source']
        )
        self._fields = [
            'name', 'topic_id', 'source_name', 'source_id', 'namespace',
            'primary_keys', 'contains_pii', 'cluster_type'
        ]
        self.source_name = topic_obj.source.name
        self.source_id = topic_obj.source.source_id
        self.namespace = topic_obj.source.namespace.name
        if kafka_topics is not None:
            self.in_kafka = self.name in kafka_topics
            self._fields.append('in_kafka')
        if topics_to_range_map is not None:
            self.message_count = self._get_topic_message_count(
                topics_to_range_map
            )
            if self.message_count:
                self.in_kafka = True
            self._fields.append('message_count')
        self._fields.extend(['created_at', 'updated_at'])

    def _get_topic_message_count(self, topics_to_range_map):
        if self.name in topics_to_range_map:
            return sum(topics_to_range_map[self.name].values())
        return 0
